/*
 * The following program reads an electronic mail message, searches its
 * database, and tries to reply. 
 *
 * To explore the program, compile with -DDEBUG.  Instead of executing shell
 * commands they will be sent to admin/stderr.  The overall context is that
 * the program is executing in a home directory, say /netlib, which contains
 * various source directories, such as /netlib/eispack.  The netlib
 * administrative directory /netlib/admin contains various indices,
 * permission tables, and logfiles. 
 *
 * This program may be freely copied. No warranty is made as to correctness or
 * suitability, but bug reports (to research!ehg) will probably receive
 * prompt attention. 
 *
 *	hacked enough that ehg may not recognize it... seismo!rick
 */

#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <sys/file.h>
#define SAME 0

/*
 * System V and Version 9 use strchr where BSD uses index. The following hack
 * tries to accommodate.  
 */
#include <sys/param.h>
#ifndef CLBYTES
#define index strchr
#define rindex strrchr
#endif
char * malloc(), *index(), *rindex();

#define LINESIZE 2000
#define FIELDMAX 400
char *FS, *field[FIELDMAX];
int NR, NF;			/* these are filled by awk() */
typedef struct hashnode {
	char *key;
	int code;
	struct hashnode *next;
} HASHNODE;
#define HASHMAX 211
typedef HASHNODE *HASHTAB[HASHMAX];

HASHNODE * hashlook(), *hash0(), *hashnext();

int group;			/* 0=general, 1=local */
int depend;			/* send all routines request depends on? */
int upas;			/* does this system run Presotto's mailer?
				 * (used in parse) */
char reply[20];			/* temp file for constructing reply */
int SEND, LIST, INDEX, FIND;	/* command codes */
#define SKIP 0

char *suffix[] = {"", ".r", ".f", ".c", 0};
/* for filename expansion in only() */

/*
 * The following subterfuge, together with the stderr redirection below, is
 * designed to append error messages of this process and its children to
 * /netlib/admin/stderr, properly interleaved. 
 */
char err_str[500];
#define error(f,p) sprintf(err_str,f,p);write(2,err_str,strlen(err_str))

#define MAILER	"/usr/lib/sendmail"
#define RETURN_ADDRESS "netlibd"
#define HOMEDIR "/usr/spool/ftp"

main()
{
	char address[LINESIZE];	/* return address */
	int cmd;		/* code for type of request in current input
				 * line */
	int item[2], excl[2], lib[2];	/* first and last field indices */
	/* for routines, exclusions, libraries */
	char zlib[500];		/* strcat of field[lib[0]] ... field[lib[1]] */
	char xlib[1000];	/* zlib after expansion through LIBS file */
	char zitem[500], zexcl[500];
	char files[LINESIZE];	/* filenames to be sent as reply */
	int nreq = 0;		/* number of recognized requests */
#ifdef TOOBIG
	int kbreq = 0;		/* number of kilobytes sent so far */
#endif
	char buf[LINESIZE];
	long size, fsiz();
	FILE *F, *popen();
	char *date, *getdate();

	chdir(HOMEDIR);	/* MODIFY AS DESCRIBED IN README. */

	/* redirect stderr */
	close(2);
	open("admin/stderr", O_WRONLY|O_APPEND);

	date = getdate();
	strcpy(reply, "/tmp/netlibXXXXXX");
	mktemp(reply);		/* removed by quit() */
	address[0] = '\0';
	upas = (access("/usr/lib/upas", 0) == 0);

	while (awk() != EOF) {	/* read next line of message */
		cmd = parse(address, item, excl, lib);
		if ((address[0] == '\0') ||
		    (cmd == SKIP) ||
		    duplicate(cmd, item, excl, lib))
			continue;

		/* translate library names */
		if (cmd != INDEX) {
			if (!lib[0])
				field[lib[0] = lib[1] = 1] = "DEFAULT";
			gather(0, lib, zlib);
			if ((cmd == SEND) || (cmd == LIST)) {
				expand(lib, xlib);
			} else {/* FIND */
				strcpy(xlib, zlib);
				if ((group == 1) && (strcasecmp(xlib, "DEFAULT") == SAME))
					strcat(xlib, " DEFAULT.lcl");
			}
			gather(depend, item, zitem);
			gather(depend, excl, zexcl);
		}
#ifdef DEBUG
		error("address=%s\n", address);
		error("cmd=%d\n", cmd);
		error("zitem=%s\n", zitem);
		error("zexcl=%s\n", zexcl);
		error("xlib=%s\n", xlib);
#endif

		F = fopen(reply, "w");
		fprintf(F, "From: %s\n", RETURN_ADDRESS);
		fprintf(F, "To: %s\n", address);
		fprintf(F, "Subject: %s\n\n", field[0]);
		fclose(F);

		/* execute command */
		if (cmd == INDEX) {
			sprintf(buf,
				group ? "cat admin/index admin/index.lcl >>%s 2>/dev/null" :
				"cat admin/index >>%s",
				reply);
			shell(buf);
		} else if (xlib[0] == '\0') {
			sprintf(buf, "cat admin/mess/sorry_lib >>%s", reply);
			shell(buf);
		} else if (unsafe(xlib) || unsafe(zitem) || unsafe(zexcl)) {
			error("unsafe! zitem=%s", zitem);
			error(" zexcl=%s", zexcl);
			error(" xlib=%s\n", xlib);
			quit(88);
		} else if (cmd == FIND) {
			sprintf(buf, "admin/bin/mgrep '%s' %s >>%s", xlib, zitem, reply);
			shell(buf);
		} else {	/* (cmd==SEND) || (cmd==LIST) */
			if (depend) {
				sprintf(buf, "admin/bin/ldMx '%s' '%s' '%s'", xlib, zitem, zexcl);
				F = popen(buf, "r");
				fgets(files, LINESIZE, F);
				pclose(F);
				nlchop(files);
				if (!*files) {	/* failed; try only() */
					depend = 0;
				}
			}
			stripl(xlib);
			if (!depend)
				only(xlib, item, files);
			nlchop(files);
			if (strncasecmp("admin/mess/sorry", files, 16) != SAME) {
				sprintf(buf, "admin/bin/disclaim %s >>%s", xlib, reply);
				shell(buf);
			}
			if (cmd == LIST) {
				size = fsiz(reply);
				F = fopen(reply, "a");
				fprintf(F, "The following files total %ld bytes\n", size);
				if (size > 90000.)
					fprintf(F, " = %.1f hours per link at 1200 baud.\n", size / 360000.);
				fprintf(F, "\n%s\n", files);
				fclose(F);
			} else {
				sprintf(buf,
				"for i in %s; do /bin/cat $i || /usr/ucb/zcat $i; done >>%s",
					files, reply);
				shell(buf);
			}
		}

		/* is user being too greedy? */
		size = fsiz(reply);
#ifdef TOOBIG
		kbreq += size / 1000;
		if ((nreq++ > 30) || (kbreq > 300)) {
			sprintf(buf, "cat admin/mess/sorry_many | %s %s", 
				MAILER, address);
			shell(buf);
			quit(5);
		}
#else /* TOOBIG */
		nreq++;
#endif /* TOOBIG */
		/* ship it out */
		if ((group == 0) && (size > 99000)) {
			sprintf(buf, "sed 1,2d %s|admin/bin/bigmail %s '%s'",
				reply, address, unsafe(field[0]) ? "your netlib request" : field[0]);
		} else {
			sprintf(buf, "%s %s <%s", MAILER, address, reply);
		}
		shell(buf);

		/* append summary to log */
		if ((F = fopen("admin/log", "a")) != NULL) {
			if (cmd == INDEX) {
				fprintf(F, "%s [%s]%s %ld INDEX\n", date, address,
					group ? "l" : "", size);
			} else {
				fprintf(F, "%s [%s]%s %ld %s / %s", date, address,
					group ? "l" : "", size, zlib, zitem);
				if (excl[0])
					fprintf(F, "~%s", zexcl);
				if (cmd == LIST)
					fprintf(F, " L\n");
				else if (cmd == FIND)
					fprintf(F, " F\n");
				else
#ifdef DEBUG
					fprintf(F, " DEBUG\n");
#else
					fprintf(F, "\n");
#endif
			}
			fclose(F);
		}
		/* continue to next request */
	}

	if (nreq == 0) {
		sprintf(buf, "cat admin/mess/sorry_req | %s %s", MAILER,
			address);
		shell(buf);
		quit(3);
	}
	quit(0);
}

shell(s)
char *s;
{
#ifdef DEBUG
	error("[%s]\n", s);
#else
	system(s);
#endif
}


only(xlib, item, files)
char *xlib;
int item[2];
char *files;
{
	/* list all existing files of the form xlib/item */
	int nlib, nfiles, i, j, k;
	char buf[LINESIZE];
	char *lib[FIELDMAX];

	nlib = split(xlib, lib, " \t");
	nfiles = 0;
	*files = '\0';
	for (i = 0; i < nlib; i++) {
		for (j = item[0]; j <= item[1]; j++) {
			for (k = 0; suffix[k]; k++) {
				sprintf(buf, "%s/%s%s", lib[i], field[j], suffix[k]);
				if (access(buf, 4) == -1){
					char zbuf[LINESIZE];
					sprintf(zbuf, "%s.Z", buf);
					if (access(zbuf, 4) == -1)
						continue;
				}

				if (nfiles)
					strcat(files, " ");
				strcat(files, buf);
				nfiles++;
				break;	/* skip remaining suffixes */
			}
		}
	}
	if (nfiles == 0) {
		if ((group == 0) && (strcasecmp(xlib, "portP") == SAME)) {
			strcpy(files, "admin/mess/sorry_port");
		} else {
			strcpy(files, "admin/mess/sorry");
			sprintf(buf, "%s/index", field[item[0]]);
			if (access(buf, 4) == 0) {
				strcat(files, " admin/mess/sorry_indiv ");
				strcat(files, buf);
			}
		}
	}
}

int
parse(address, item, excl, lib)
char address[];
int item[2], excl[2], lib[2];
{
	/* analyze field[] */
	int cmd, k;
	char c, *p;
	static char buf[LINESIZE];
	static int in_From = 0;
	static int in_header = 1;
	static int first = 1;
	static HASHTAB keywords;
	static int FROM, FROMC, ONLY, OF, WHO, IS, WHOIS, BUT, NOT;
	static int SUBROUTINE, IGNORE;
	int code[FIELDMAX];	/* = lookup(field[],keywords); */
#ifdef DEBUG
	register int i;
#endif /* DEBUG */
	FILE *F;

	if (first) {
		first = 0;
		k = SKIP;
#define keyword(s,i) insert(s,i = ++k,keywords)
		keyword("From", FROM);
		keyword("From:", FROMC);
		keyword("send", SEND);
		keyword("list", LIST);
		keyword("only", ONLY);
		keyword("of", OF);
		keyword("whois", WHOIS);
		keyword("who", WHO);
		keyword("is", IS);
		keyword("find", FIND);
		keyword("Subject:", IGNORE);
		keyword("index", INDEX);
		keyword("but", BUT);
		keyword("not", NOT);
		keyword("subroutine", SUBROUTINE);
#define synonym(s,i) insert(s,i,keywords)
		synonym("Override-From:", FROM);	/* These allow a
							 * knowlegable user */
		synonym("Really-From:", FROM);	/* to bypass wrong mail
						 * transformers. */
		synonym("please", IGNORE);
		synonym("sned", SEND);
		synonym("mail", SEND);
		synonym("request", SEND);
		synonym("envoyer", SEND);
		synonym("get", SEND);
		synonym("from", OF);
		synonym("<", OF);
		synonym("for", OF);
		synonym("4", OF);
		synonym("in", OF);
		synonym("where", WHO);
		synonym("directory", INDEX);
		synonym("information", INDEX);
		synonym("info", INDEX);
		synonym("help", INDEX);
		synonym("subroutines", SUBROUTINE);
		synonym("routine", SUBROUTINE);
		synonym("routines", SUBROUTINE);
	}
	/*
	 * Beware.  Extracting the return address is the most delicate part
	 * of netlib.  Don't tweak unless you know what you are doing.  
	 */
	if (in_From) {		/* continuation of From: line */
		c = *field[0];
		if (c == ' ' || c == '\t') {
			if (strlen(field[0]) + strlen(buf) >= LINESIZE)
				quit(2);
			strcat(buf, field[0]);
			return (SKIP);
		} else {	/* not a continuation */
			in_From = 0;
			if (index(buf, '(') || index(buf, '<') || index(buf, '[')) {
				/*
				 * address appears to contain actual as well
				 * as electronic name 
				 */
				if ((F = fopen("admin/names", "a")) != NULL) {
					fprintf(F, "%s\n", buf);	/* log it */
					fclose(F);
				}
			}
			if (!upas) {	/* Experience shows that with the
					 * upas mailer, From lines are more
					 * reliable than From: lines.  Under
					 * other mailers, the From line is
					 * frequently scrambled. */
				if (p = index(buf, ','))
					*p = '\0';	/* chop all but first
							 * author */
				if (getfrom(buf, address) != 1)
					quit(3);	/* extract RFC822
							 * address */
				chkaddr(address);
			}
		}
	}
	if (NF == 0) {		/* blank line */
		in_header = 0;
		return (SKIP);
	}
	code[1] = lookup(field[1], keywords);

	/* look for (start of) return address */
	if (in_header && (NF > 1)) {
		if (code[1] == FROM) {
			strcpy(address, field[2]);
			chkaddr(address);
			return (SKIP);
		} else if (code[1] == FROMC) {
			strcpy(buf, 6 + field[0]);
			in_From = 1;
			return (SKIP);
		}
	}			/* ignore bogus From: lines added by IBM mail
				 * program */
	/*  ignore anything after ". " or " #" */
	for (k = 1; k <= NF; k++) {
		p = field[k] + strlen(field[k]) - 1;
		if (*p == '.') {
			*p = '\0';
			NF = k;
		} else if (*field[k] == '#') {
			NF = k - 1;
		}
	}

	/* allow MS-DOS syntax */
	for (p = field[0]; p = index(p, '\\');)
		*p = '/';

	/* finish decoding */
	for (k = 2; k <= NF; k++)
		code[k] = lookup(field[k], keywords);
	for (k = 1; (code[k] == IGNORE) && (++k <= NF);) {
	}
	if (k > NF)
		return (SKIP);
#ifdef DEBUG
	error("%s", "\n");
	for (i = k; i <= NF; i++) {
		error("%s", field[i]);
		error("=%d ", code[i]);
	}
	error("%s", "\n");
#endif

	/* now look for a request */
	depend = 1;
	cmd = code[k];
	if ((cmd == SEND) || (cmd == LIST)) {
		if ((k + 1 <= NF) && (code[k + 1] == ONLY)) {
			k++;
			depend = 0;
		}
		if ((k + 2 <= NF) && (code[k + 1] == LIST) && (code[k + 2] == OF)) {
			k += 2;
			cmd = LIST;
		} else if ((k + 1 == NF) && (code[k + 1] == INDEX)) {
			k += 1;
			cmd = INDEX;
		}
	} else if ((cmd == FIND) || (cmd == WHOIS)) {
		depend = 0;
	} else if (cmd == WHO) {
		if ((k + 2 <= NF) && (code[k + 1] == IS)) {
			k += 1;
			cmd = WHOIS;
			depend = 0;
		} else {
			cmd = SKIP;
		}
	} else if (cmd != INDEX) {
		cmd = SKIP;
	}
	k++;			/* now field[k] is first item requested */

	if ((cmd == SKIP) || (cmd == INDEX))
		return (cmd);

	/* get items, exclusions, libraries */
	excl[0] = lib[0] = 0;
	excl[1] = lib[1] = -1;
	if ((k + 1 <= NF) && code[k] == SUBROUTINE)
		k++;
	item[0] = k;
	for (; k <= NF; k++) {
		if ((k + 1 <= NF) && (code[k] == BUT) && (code[k + 1] == NOT)) {
			item[1] = k - 1;
			k = k + 2;
			excl[0] = k;
		} else if (code[k] == OF) {
			if (excl[0])
				excl[1] = k - 1;
			else
				item[1] = k - 1;
			lib[0] = ++k;
		}
	}
	if (lib[0])
		lib[1] = k - 1;
	else if (excl[0])
		excl[1] = k - 1;
	else
		item[1] = k - 1;

	if ((cmd == SEND) || (cmd == LIST)) {
		if (depend && (item[0] == item[1]) && (code[item[0]] == INDEX))
			depend = 0;
	} else if (cmd == WHOIS) {
		cmd = FIND;
		if (!lib[0]) {
			assert(item[1] < FIELDMAX - 2);
			k = FIELDMAX - 2;
			field[k] = "golub";
			lib[0] = k;
			lib[1] = k;
		}
	}
	return (cmd);
}

int
duplicate(cmd, item, excl, lib)
int cmd, item[2], excl[2], lib[2];
{
	/* Is the current request the same as the previous one? */
	static int execnt = 0;
	static int c, i[2], e[2], l[2];
	static char line[LINESIZE], *f[FIELDMAX];
	char *next, *cpfld();

	if ((execnt != 0) && (c == cmd) &&
	    (cmpfld(i, f, item, field) == SAME) &&
	    (cmpfld(e, f, excl, field) == SAME) &&
	    (cmpfld(l, f, lib, field) == SAME)) {
		return (1);
	}
	execnt++;

	/* save current request */
	c = cmd;
	i[0] = item[0];
	i[1] = item[1];
	e[0] = excl[0];
	e[1] = excl[1];
	l[0] = lib[0];
	l[1] = lib[1];
	next = cpfld(line, i, f);
	next = cpfld(next, e, f);
	next = cpfld(next, l, f);
	return (0);
}

int
cmpfld(i, f, j, g)
int i[2], j[2];
register char *f[], *g[];
{
	register int k = i[1] - i[0];
	if (k != (j[1] - j[0]))
		return 1;
	while (k >= 0) {
		if (strcasecmp(f[i[0] + k], g[j[0] + k]) != SAME)
			return 1;
		k--;
	}
	return SAME;
}

char *
cpfld(p, i, f)
register char *p;			/* next position to put chars */
int i[2];
char *f[];
{
	register int k;
	for (k = i[0]; k <= i[1]; k++) {
		f[k] = p;
		strcpy(p, field[k]);
		p += strlen(p) + 1;
	}
	return p;
}

/********* expand library names ***********/

expand(lib, xlib)
int lib[2];
char *xlib;
{
	register int k;

	xlib[0] = '\0';
	for (k = lib[0]; k <= lib[1]; k++) {
		if (!group || !search("admin/LIBS.lcl", field[k], xlib))
			search("admin/LIBS", field[k], xlib);
	}
	if (depend && (strncasecmp("-l", xlib, 2) != SAME))
		depend = 0;
	if (!depend)
		stripl(xlib);
}

int				/* returns 1 if found, 0 if not */
search(filename, pat, xlib)
char *filename;
char *pat;
char *xlib;
{
	/* find first line in file starting with pat */
	FILE *f;
	int len, strlen(), matches = 0;
	char line[BUFSIZ];

	if (*filename == '\0')
		return 0;
	if ((f = fopen(filename, "r")) == NULL)
		return 0;
	len = strlen(pat);
	while (fgets(line, BUFSIZ, f) != NULL) {
		nlchop(line);
		if ((strncasecmp(line, pat, len) == SAME) && (line[len] == ' ')) {
			found(line, xlib);
			matches++;
			break;
		}
	}
	fclose(f);
	return matches;
}

found(s, xlib)
char *s;
char *xlib;
{
	char *r, *index();
	r = index(s, '>');	/* strip "^.*=> *" */
	if (r == 0) {
		error("corrupted LIBS? %s", s);
		quit(1);
	}
	r++;
	while (isspace(*r))
		r++;
	if (xlib[0] != '\0')
		strcat(xlib, " ");
	strcat(xlib, r);
}

stripl(r)
char *r;
{
	/* remove, inplace, occurrences of "-l" */
	char *p;
	if (*r == '\0')
		return;
	for (p = r + 1; *p; p++) {
		if ((*r == '-') && (*p == 'l')) {
			r--;
		} else {
			*++r = *p;
		}
	}
	*++r = '\0';
}

/******* netlib-specific symbol table *********/

insert(s, i, tab)
char *s;
int i;
HASHTAB tab;
{
	HASHNODE *p = (HASHNODE *) malloc(sizeof(HASHNODE));
	assert(p);
	p->key = s;
	p->code = i;
	hashins(p, tab);
}

int
lookup(s, tab)
char *s;
HASHTAB tab;
{
	/* strips trailing punctuation and may convert to lower case */
	/* also strips surrounding brackets */
	HASHNODE *p;
	register char *r = s + strlen(s) - 1;	/* end of string */
	register char *t;
	if (((*s == '<') && (*r == '>')) ||
	    ((*s == '[') && (*r == ']')) ||
	    ((*s == '{') && (*r == '}'))) {
		*r = '\0';
		for (t = s; t < r; t++)
			t[0] = t[1];
	}
	for (; (r >= s) && index(",?;\"'`.", *r); r--) {
	}
	r[1] = '\0';
	p = hashlook(s, tab);
	if (!p && casech(s))
		p = hashlook(s, tab);
	return (p ? p->code : 0);
}

int
casech(s)
register char *s;
{
	/* return non-zero if case changed to lower */
	register int changed = 0;
	while (*s) {
		if (isupper(*s)) {
			*s = tolower(*s);
			changed = 1;
		}
		s++;
	}
	return (changed);
}

handle(line)
char *line;
{
	/* save message, toss escapes */
	register char *s;	/* next char to examine */
	register char *t;	/* next place to copy char */
	static int execnt = 0;
	static FILE *save;
	if (execnt++ == 0)
		save = fopen("/tmp/netreq", "a");
	/* closed automatically when program quits */
	/*
	 * This won't work if several netlib processes are running
	 * simultaneously;  but this log isn't too important, so we won't
	 * worry about that. 
	 */
	fprintf(save, "%s\n", line);
	/* delete nonprinting characters; interpret backspace and tab */
	for (s = line, t = line; *s; s++) {
		if ((*s == '\b') && (t > line)) {
			t--;	/* might not be right for ESC BS, but tough */
		} else if (*s == '\t') {
			*t++ = ' ';
		} else if (!iscntrl(*s)) {
			*t++ = *s;
		}
	}
}

/*************** utilities modeled after awk ****************/
static char line[LINESIZE], line2[LINESIZE];
static char isFS[256], isntFS[256], *oFS;

int
awk()
{
	if (gets(field[0] = line2) == NULL)
		return EOF;
	++NR;
	handle(line2);
	strcpy(line, field[0]);
	if (!FS)
		FS = " \t";
	NF = split(line, field + 1, FS);
	return 1;
}

int
split(s, a, r)
register unsigned char *s, *a[], *r;
{
	register int i = 0;

	if (!oFS || strcasecmp(r, oFS))
		FSinit();

	while (i < FIELDMAX - 1) {
		while (isFS[*s])
			s++;
		a[i++] = s;
		while (isntFS[*s])
			s++;
		if (*s == '\0')
			break;
		*s++ = '\0';
	}
	if (!*a[i - 1])
		i--;
	return i;
}

FSinit()
{
	register int i;
	register unsigned char *s;
	for (i = 0; i < (sizeof isFS); i++) {
		isFS[i] = 0;
		isntFS[i] = 1;
	}
	for (s = (unsigned char *) FS; *s; s++) {
		isFS[*s] = 1;
		isntFS[*s] = 0;
	}
	isntFS[0] = 0;
	oFS = FS;
}

/*
 * Gentle Reader:  I realize that hashing is overkill for this application. 
 * But the code was already lying around and you might find it useful
 * elsewhere.  
 */
int hash();

HASHNODE *
hashlook(key, tab)
char *key;
HASHTAB tab;
{
	int i = hash(key);
	HASHNODE *p = tab[i];
	HASHNODE *q;		/* q->next == p */
	if (!p)
		return (0);
	if (strcasecmp(key, p->key) == 0)
		return (p);
	q = p;
	while (p = p->next) {
		if (strcasecmp(key, p->key) == 0) {
			/* move to front */
			q->next = p->next;
			p->next = tab[i];
			tab[i] = p;
			return (p);
		}
		q = p;
	}
	return (0);
}

hashins(n, tab)
HASHNODE *n;
HASHTAB tab;
{
	HASHNODE **pp = tab + hash(n->key);
	n->next = *pp;
	*pp = n;
}

hashdel(n, tab)
HASHNODE *n;
HASHTAB tab;
{
	HASHNODE **pp = tab + hash(n->key);
	while (*pp) {
		if (*pp == n) {
			*pp = n->next;
			n->next = 0;	/* to limit disasters */
			return;
		}
		pp = &((*pp)->next);
	}
	error("hashdel: couldn't find node %s", n->key);
	quit(1);
}

HASHNODE *
hash0(tab)
HASHTAB tab;
{
	int i = 0;
	while (tab[i] == 0)
		if ((++i) >= HASHMAX)
			return (0);
	return (tab[i]);
}

HASHNODE *
hashnext(p, tab)
HASHNODE *p;
HASHTAB tab;
{
	int i;
	assert(p);
	if (p->next)
		return (p->next);
	i = 1 + hash(p->key);
	if (i >= HASHMAX)
		return (0);
	while (tab[i] == 0)
		if ((++i) >= HASHMAX)
			return (0);
	return (tab[i]);
}

int
hash(s)
char *s;
{
	/* hashpjw from Aho+Sethi+Ullman */
	char *p;
	unsigned h = 0;
	unsigned g;
	for (p = s; *p; p++) {
		h = (h << 4) + *p;
		if (g = h & 0xf0000000) {
			h = h ^ (g >> 24);
			h = h ^ g;
		}
	}
	return (h % HASHMAX);
}


/************* return address nonsense *****************/

chkaddr(address)
char address[];
{
	int n, groupid();
	char *p, addr2[LINESIZE];

	if (!*address || unsafe(address))
		quit(99);
	group = groupid(address);
	if (group >= 10)
		quit(11);	/* enemy! Don't even give a reply. */
	assert(group >= 0);
	assert(group <= 1);

	/*
	 * perhaps the following should be just replaced by a search for the
	 * string "netlib" somewhere in the address.  
	 */
	n = strlen(address);
	if ((n >= 6) && strcasecmp("netlib", address + n - 6) == SAME) {
		error("infinite loop?  address=%s\n", address);
		quit(7);
	}
	if (group == 1) {	/* if local, chop to single hop return */
		strcpy(addr2, address);
		if (p = rindex(addr2 + 1, '!')) {	/* addr2 is certainly
							 * nonempty */
			p--;
			while ((p != addr2) && (*p != '!')) {
				p--;
			}
			if (*p == '!')
				strcpy(address, p + 1);
		}
	}
}


/*
 * This returns the "security clearance" of address. Uucp-style address
 * syntax "machine!...!user" is assumed. return codes: 0 = general address 1
 * = local  (all machines are in file "local") 10 = enemy  (some machine is
 * in file "enemies") 
 */

int
groupid(a)
char *a;
{
	char *m[100];		/* machines in address */
	int n = 0;		/* size of m */
	char *p, *q;
	int cmp();

	p = malloc(strlen(a) + 1);
	assert(p);
	strcpy(p, a);

	/* strip off username (doesn't work for person@host style addresses) */
	q = rindex(p, '!');
	if (q != NULL) {
		*q = '\0';
		if (strcasecmp(q + 1, "uucp") == 0)
			return (10);
	} else {
		q = rindex(p, '@');
		if (q == NULL)
			return (1);
		strcpy(p, q+1);
	}

	/* split at '!' and '@' */
	m[n++] = p;
	while (*++p) {
		if (*p == '!' || *p == '@') {
			*p++ = '\0';
			m[n++] = p;
		}
	}
	qsort(m, n, sizeof(m + 0), cmp);

	if (del(m, &n, "admin/groups/enemies"))
		return (10);
	del(m, &n, "admin/groups/local");
	if (n == 0)
		return (1);
	return (0);
	/* won't be many addresses, so don't bother to free */
}

int
cmp(p, q)
char **p, **q;
{
	/* needed for qsort */
	return (strcasecmp(*p, *q));
}

int
del(m, nn, file)
char *m[];
int *nn;
char *file;
{
	/* deletes entries in m that occur in file;  returns number deleted */
	/* file and m are assumed to be sorted. */
	register int n = *nn;
	register int i;		/* walks through m */
	int j;			/* continue place to put item in m;  if j==i,
				 * no need to copy */
	int ndel = 0;		/* number of deletions */
	int c;			/* result of string comparison */
	char a[100];	/* machine name from file */
	FILE *fopen(), *f = fopen(file, "r");
	if (f == NULL)
		return 0;
	i = j = ndel = 0;
	if (fgets(a, 100, f) == NULL)
		return 0;
	nlchop(a);
	while (1) {
		c = strcasecmp(a, m[i]);
		if (c < 0) {
			/* advance a */
			if (fgets(a, 100, f) == NULL)
				break;
			nlchop(a);
		} else if (c > 0) {
			if (i > j)
				m[j] = m[i];
			j++;
			i++;
			if (i >= n)
				break;	/* advance m */
			continue;
		} else {	/* match */
			ndel++;
			i++;
			if (i >= n)
				break;	/* advance m */
			continue;
		}
	}
	fclose(f);
	if (i > j)
		while (i < n)
			m[j++] = m[i++];
	*nn = n - ndel;
	return (ndel);
}

/******* other utilities *********/

#include <time.h>		/* on some systems, this is <sys/times.h> */
char datestr[30];

char *
getdate()
{
	struct tm *tp, *localtime();
	long clock, time();
	clock = time(0);
	tp = localtime(&clock);
	sprintf(datestr, "%d %2d %2d %02d%02d", tp->tm_year + 1900, tp->tm_mon + 1,
		tp->tm_mday, tp->tm_hour, tp->tm_min);
	return datestr;
}

quit(rc)
int rc;
{
	unlink(reply);
	if (rc != 0) {
		error("%s exit", datestr);
		error(" %d\n", rc);
	}
	exit(0);		/* some mail programs demand 0 return from
				 * alias script */
}

nlchop(s)
register char *s;
{
	/* chop trailing newline */
	if (!*s)
		return;
	s += strlen(s) - 1;
	if (*s == '\n')
		*s = '\0';
}

gather(chop, f, z)
int chop, f[2];
char z[];
{
	/* catenate fields f[0]...f[1], possibly chopping suffix */
	int k;
	z[0] = '\0';
	for (k = f[0]; k <= f[1]; k++) {
		if (chop)
			suffchop(field[k]);
		if (k > f[0])
			strcat(z, " ");
		strcat(z, field[k]);
	}
}

suffchop(s)
register char *s;
{
	/* chop suffix */
	if (!*s)
		return;
	if (s = rindex(s, '.'))
		*s = '\0';
}

long
fsiz(s)
char *s;
{
	int fd;
	long t;
	long lseek();
	if ((fd = open(s, 0)) == -1) {
		error("can't open %s\n", s);
		quit(7);
	}
	if ((t = lseek(fd, 0L, 2)) == -1) {
		error("can't seek on %s\n", s);
		quit(8);
	}
	close(fd);
	return t;
}
